/* 
 * FILE --- rk4_sim2.hh
 *
 * PURPOSE --- Header file for a class for a 4'th order Runge-Kutta 
 * method for solving a system of  n first order differential equations. 
 * If the initial equation is in the form of n'th order differential equation
 * it must be converted to a system of n first order differential
 * equations.
 *
 */

// Prevent Multiple Inclusion
#ifndef RK4_SIM2_HH
#define RK4_SIM2_HH

// Include Files
#include <stdlib.h>

// Local Definitions
#define RK4_SIM2_NO_ERROR 0
#define RK4_SIM2_MEM_ERROR 1
#define RK4_SIM2_UNKNOWN_IC_ERROR 2


class RK4_SIM2
{
public:
  /*
   Constructor
   "deriv" should be a pointer to a function that contains the dynamic 
   equations of the system in the form: xdot = f(t,x,u); arguments are 
   the time t, present state values x, the present control value u, 
   and values of the derivatives of the states xdot (calculated inside of 
   the funtcion). This function should be written by the user.

   The other parameters are self-explanatory.
   If you specify initial conditions, the parameter _initial_state 
   should point to a vector of size _plant_order, so that correct 
   initialization could be performed.
  */
  RK4_SIM2(void (*_deriv) (const double t, const double *x, 
			   const double *u, double *xdot),
	   int _plant_order, 
	   int _ctrl_order, 
	   double _integration_step = 0.0001, 
	   double _initial_time = 0.0, 
	   double *_initial_state = NULL);  
  
  // Destructor
  ~RK4_SIM2();
  
  double *plant_state; // Current System states
  double *ctrl_input; // Current Controller output
  double time;  // Current time
  
  int rk4_sim_err;   // Variable to hold an error number
  
  void solve (void); // Performs one step simulation
  
private:
  int plant_order; // Num of plant states
  int ctrl_order; // Num of control inputs
  double integration_step; // Integration step size
  
  double *f1, *f2, *f3, *f4, *temp; // Runge-Kutta Coefficients
  
  void allocate_memory (void);
  
  inline void F1 (void); // Functions which calculate coefficients
  inline void F2 (void);
  inline void F3 (void);
  inline void F4 (void);
  
  void (*deriv) (const double t, const double *x, 
		 const double *u, double *xdot);
};

#endif // RK4_SIM2_HH


